﻿Type=StaticCode
Version=7.01
ModulesStructureVersion=1
B4A=true
@EndOfDesignText@
'************************************************************************************
'
'This code module calculates V4 signatures for Amazon Web Services (AWS) Simple 
'Storage Service (S3) REST API requests, enabling creation of AWS S3 PUT/GET/DELETE 
'requests, for introduction see:
'
'http://docs.aws.amazon.com/AmazonS3/latest/API/Welcome.html
'
'Following characteristics/restrictions are adopted:
'
'   o Authentication with AWS Signature version 4 using a HTTP Authorization header,
'     this dictates appropriate request authentication guide is via:
'
'     http://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-auth-using-authorization-header.html
' 
'   o Payloads passed as a single chunk, this dictates appropriate request 
'     authentication guide is via:
'
'     http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
' 
'Calls:
'
'   Signature
'
'   HexSHA256Hash
'
'   HMACSHA256
'
'Properties:
'
'	AWS_Access_Key_ID
'	AWS_Secret_Access_Key
'	AWS_Region
'       where S3 bucket resides, as per:
'       http://docs.aws.amazon.com/general/latest/gr/rande.html#s3_region
'       e.g.:
'           ap-southeast-2
'	AWS_S3_Bucket_Name
'       when URI is "path style", e.g.:
'           http://<endpoint>/<bucketname>/<filename>
'       specify AWS_S3_Bucket_Name
'       when URI is "virtual hosted style", e.g.:
'           http://<bucketname>/<endpoint>/<filename>
'       specify AWS_S3_Bucket_Name of ""
'	AWS_S3_File_Name
'       full name of file in bucket, including any pseudo folders, e.g.:
'           examplefile.txt
'           pseudofolder1/pseudofolder2/examplefile.txt
'	AWS_S3_HttpMethod
'       "GET", "PUT" or "DELETE"
'	AWS_S3_Query_map
'       map of URI query parameter key, value pairs
'       e.g:
'            URI is:
'                http://s3.amazonaws.com/examplebucket?prefix=somePrefix&marker=someMarker&Max-keys=20
'            URI query parameter key, value pairs:
'                "prefix", "somePrefix"
'                "marker", "someMarker"
'                "max-keys", "20"
'       e.g:
'            URI is:
'                http://s3.amazonaws.com/examplebucket?acl
'            URI query parameter key, value pairs:
'                "acl", ""
'       e.g:
'            URI does not include a "?"
'                AWS_S3_Query_map is Null
'	AWS_S3_Payload
'	AWS_S3_GMT_DateTime
'	AWS_S3_Header_map
'       map of request headers key, value pairs - must include following headers, all
'       others are optional:
'             Host header
'             Content-Type header - if present in request
'             Any x-amz-* headers
'       consider including all headers in order to prevent data tampering
'
'Requirements:
'
'   o B4A libraries:
'
'         ByteConverter (version 1.10 or later)
'         Encryption (version 1.10 or later)
'
'Update history:
'
'   24 Jun 17 - 1.0
'
'************************************************************************************
Private Sub Process_Globals
	
	Public AWS_Access_Key_ID As String
	Public AWS_Secret_Access_Key As String
	Public AWS_Region As String                        

	Public AWS_S3_Bucket_Name As String
	Public AWS_S3_File_Name As String
	Public AWS_S3_HttpMethod As String
	Public AWS_S3_Query_map As Map
	Public AWS_S3_Payload As String
	Public AWS_S3_Header_map As Map
	Public AWS_S3_GMT_DateTime As Long
	
	Type KeyValueType(key As String, value As String)

End Sub

'************************************************************************************
'
'This procedure creates a CanonicalRequest as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 1: Create a Canonical Request
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       CanonicalRequest
'
'Notes on this procedure:
'
'       <HTTPMethod>\n
'       <CanonicalURI>\n
'       <CanonicalQueryString>\n
'       <CanonicalHeaders>\n
'       <SignedHeaders>\n
'       <HashedPayload>'
'
'************************************************************************************
Private Sub CanonicalRequest As String
	
	Private wrk_str As String
	
	'Components are separated by \n which is Chr(10) which is CRLF
	
	wrk_str = AWS_S3_HttpMethod & CRLF
	
	wrk_str = wrk_str & CanonicalURI & CRLF
	
	wrk_str = wrk_str & CanonicalQueryString & CRLF
	
	wrk_str = wrk_str & CanonicalHeaders & CRLF
	
	wrk_str = wrk_str & SignedHeaders & CRLF
	
	wrk_str = wrk_str & HexSHA256Hash(AWS_S3_Payload)

	Return wrk_str

End Sub

'************************************************************************************
'
'This procedure creates a CanonicalURI as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 1: Create a Canonical Request
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       CanonicalURI
'
'Notes on this procedure:
'
'       o CanonicalURI is URI-encoded ("/" excluded) version of absolute path 
'         component of URI — everything starting with "/" that follows domain name 
'         and up to end of string or to question mark character ("?") if there are 
'         query string parameters
'       o Example:
'            URI is (path style):
'                http://s3-ap-southeast-2.amazonaws.com/mybucket/myfolder/myphoto.jpg
'            absolute path is:
'                /mybucket/myfolder/myphoto.jpg
'       o Example:
'            URI is (virtual hosted style):
'                http://mybucket/s3-ap-southeast-2.amazonaws.com/myfolder/myphoto.jpg
'            absolute path is:
'                /myfolder/myphoto.jpg
'
'************************************************************************************
Private Sub CanonicalURI As String
	
	Private wrk_str As String
	
	'Build absolute path component of URI
	
	wrk_str = "/"
	
	'If bucket name exists, URI must be path style, add bucket name to sork string
	If AWS_S3_Bucket_Name <> "" Then
		wrk_str = wrk_str & AWS_S3_Bucket_Name & "/"
	End If
	
	'If file name exists, add it to work string
	If AWS_S3_File_Name <> "" Then
		wrk_str = wrk_str & AWS_S3_File_Name
	End If
	
	'URI-encode ("/" excluded)
	wrk_str = UriEncode(wrk_str, False)
	
	Return wrk_str

End Sub

'************************************************************************************
'
'This procedure creates a CanonicalQueryString as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 1: Create a Canonical Request
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       CanonicalQueryString
'
'Notes on this procedure:
'
'       o CanonicalQueryString specifies URI-encoded query string parameters -
'         parameter key names and values are URI-encoded individually then
'         parameters are sorted alphabetically by key name
'       o Example:
'            URI is:
'                http://s3.amazonaws.com/examplebucket?prefix=somePrefix&marker=someMarker&Max-keys=20
'            canonical query string is (line breaks are added for readability):
'                URI-encode("marker") + "=" + URI-encode("someMarker") + "&" +
'                URI-encode("max-keys") + "=" + URI-encode("20") + "&" +
'                URI-encode("prefix") + "=" + URI-encode("somePrefix")
'       o When a request targets a subresource, corresponding query parameter value
'         will be an empty string (""), for example:
'            URI is:
'                http://s3.amazonaws.com/examplebucket?acl 
'            canonical query string is:
'                URI-encode("acl") + "=" + ""
'       o If URI does not include a "?", there is no query string in request, and 
'         canonical query string is:
'                ""
'
'************************************************************************************
Private Sub CanonicalQueryString As String
	
	Private wrk_ptr As Int
	Private wrk_str As String
	Private wrk_map As Map
	Private wrk_key As String
	Private wrk_value As String
	
	wrk_map.Initialize
	
	'For each key, value pair of query string parameters map...
	For wrk_ptr = 0 To AWS_S3_Query_map.Size - 1
		
		'Get key and URI-encode it
		wrk_key = AWS_S3_Query_map.GetKeyAt(wrk_ptr)
		wrk_key = UriEncode(wrk_key, True)
		
		'Get value and URI-encode it
		wrk_value = AWS_S3_Query_map.GetValueAt(wrk_ptr)
		wrk_value = UriEncode(wrk_value, True)
		
		'Save URI-encoded keys and values in temporary map
		wrk_map.Put(wrk_key, wrk_value)
		
	Next
	
	'Sort temporary map
	wrk_map = SortMap(wrk_map)
	
	'For each URI-encoded key, value pair of temporary map...
	For wrk_ptr = 0 To wrk_map.Size - 1
		
		'Append & to work string if it is not blank
		If wrk_str <> "" Then wrk_str = wrk_str & "&"
		
		'Append key=value to work string
		wrk_str = wrk_str & wrk_map.GetKeyAt(wrk_ptr) & "=" & wrk_map.GetValueAt(wrk_ptr)
		
	Next
	
	Return wrk_str

End Sub

'************************************************************************************
'
'This procedure creates a CanonicalHeaders as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 1: Create a Canonical Request
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       CanonicalHeaders
'
'Notes on this procedure:
'
'       o CanonicalHeaders is a list of request headers with their values - 
'         individual header name and value pairs are separated by newline 
'         character ("\n") i.e. CRLF - header names must be in lowercase and must 
'         be sorted alphabetically as indicated here:
'
'             Lowercase(<HeaderName1>)+":"+Trim(<value>)+"\n"
'             Lowercase(<HeaderName2>)+":"+Trim(<value>)+"\n"
'             ...
'             Lowercase(<HeaderNameN>)+":"+Trim(<value>)+"\n"
'
'       o CanonicalHeaders must include following headers, all others are optional:
'
'             Host header
'             Content-Type header - if present in request
'             Any x-amz-* headers - e.g. if temporary security credentials are
'                 used then request will include x-amz-security-token header
'
'       o Consider including all headers in order to prevent data tampering
'       o x-amz-content-sha256 header is required for all AWS Signature Version 4 
'         requests - it provides a hash of request payload - if there is no payload
'         then provide hash of an empty string
'       o Example (note sorted lowercase header names and trimmed values:
'            host:s3.amazonaws.com\n
'            x-amz-content-sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855\n
'            x-amz-date:20130708T220855Z
'
'************************************************************************************
Private Sub CanonicalHeaders As String
	
	Private wrk_ptr As Int
	Private wrk_str As String
	Private wrk_map As Map
	Private wrk_key As String
	Private wrk_value As String
	
	wrk_map.Initialize
	
	wrk_map = AWS_S3_Header_map
	
	'Sort temporary map
	wrk_map = SortMap(wrk_map)
	
	'For each key, value pair of temporary map...
	For wrk_ptr = 0 To wrk_map.Size - 1
		
		'Get key and lowercase it
		wrk_key = wrk_map.GetKeyAt(wrk_ptr)
		wrk_key = wrk_key.ToLowerCase
		
		'Get value and trim it
		wrk_value = wrk_map.GetValueAt(wrk_ptr)
		wrk_value = wrk_value.Trim
		
		'Append key:value to work string
		wrk_str = wrk_str & wrk_key & ":" & wrk_value
		
		'Append \n (CRLF) to work string
		wrk_str = wrk_str & CRLF
		
	Next
	
	Return wrk_str

End Sub

'************************************************************************************
'
'This procedure creates a SignedHeaders as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 1: Create a Canonical Request
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       SignedHeaders
'
'Notes on this procedure:
'
'       o SignedHeaders is an alphabetically sorted, semicolon-separated list of 
'         lowercase request header names - same headers that are included in 
'         CanonicalHeaders string
'       o Example (referring to example in CanonicalHeaders) equivalent SignedHeaders
'         would be:
'            host;x-amz-content-sha256;x-amz-date
'
'************************************************************************************
Private Sub SignedHeaders As String
	
	Private wrk_ptr As Int
	Private wrk_str As String
	Private wrk_map As Map
	Private wrk_key As String
	
	wrk_map.Initialize
	
	wrk_map = AWS_S3_Header_map
	
	'Sort temporary map
	wrk_map = SortMap(wrk_map)
	
	'For each key, value pair of temporary map...
	For wrk_ptr = 0 To wrk_map.Size - 1
		
		'Append ; to work string if it is not blank
		If wrk_str <> "" Then wrk_str = wrk_str & ";"
		
		'Get key and lowercase it
		wrk_key = wrk_map.GetKeyAt(wrk_ptr)
		wrk_key = wrk_key.ToLowerCase
		
		'Append key to work string
		wrk_str = wrk_str & wrk_key
		
	Next
	
	Return wrk_str

End Sub

'************************************************************************************
'
'This procedure creates a StringToSign as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 2: Create a String to Sign
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       StringToSign
'
'Notes on this procedure:
'
'       "AWS4-HMAC-SHA256" + "\n" +
'       timeStampISO8601Format + "\n" +
'       <Scope> + "\n" +
'       Hex(SHA256Hash(<CanonicalRequest>))
'
'************************************************************************************
Private Sub StringToSign As String
	
	Private wrk_str As String
	
	'Components are separated by \n which is Chr(10) which is CRLF
	
	wrk_str = "AWS4-HMAC-SHA256" & CRLF
	
	wrk_str = wrk_str & timeStampISO8601Format & CRLF
	
	wrk_str = wrk_str & Scope & CRLF
	
	wrk_str = wrk_str & HexSHA256Hash(CanonicalRequest)

	Return wrk_str

End Sub

'************************************************************************************
'
'This procedure creates a timeStampISO8601Format as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 2: Create a String to Sign
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       timeStampISO8601Format
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************
Private Sub timeStampISO8601Format As String
	
	DateTime.DateFormat = "yyyyMMdd'T'HHmmss'Z'"
	
	Return DateTime.Date(AWS_S3_GMT_DateTime)

End Sub

'************************************************************************************
'
'This procedure creates a Scope as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 2: Create a String to Sign
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       Scope
'
'Notes on this procedure:
'
'       date.Format(<YYYYMMDD>) + "/" + <region> + "/" + <service> + "/aws4_request"
'
'************************************************************************************
Private Sub Scope As String
	
	Private wrk_str As String
	
	DateTime.DateFormat = "yyyyMMdd"
	
	wrk_str = DateTime.Date(AWS_S3_GMT_DateTime) & "/"
	
	wrk_str = wrk_str & AWS_Region & "/"
	
	wrk_str = wrk_str & "s3" & "/aws4_request"
	
	Return wrk_str
 
End Sub

'************************************************************************************
'
'This procedure creates a SigningKey as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 3: Calculate Signature
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       SigningKey
'
'Notes on this procedure:
'
'       SigningKey           = HMAC-SHA256(<DateRegionServiceKey>, "aws4_request")'
'       DateRegionServiceKey = HMAC-SHA256(<DateRegionKey>, "<aws-service>")
'       DateRegionKey        = HMAC-SHA256(<DateKey>, "<aws-region>")
'       DateKey              = HMAC-SHA256("AWS4"+"<SecretAccessKey>", "<YYYYMMDD>")
'
'************************************************************************************
Private Sub SigningKey As String
	
	Return HMACSHA256(DateRegionServiceKey, "aws4_request", True)

End Sub

Private Sub DateRegionServiceKey As String
	
	Return HMACSHA256(DateRegionKey, "s3", True)

End Sub

Private Sub DateRegionKey As String
	
	Return HMACSHA256(DateKey, AWS_Region, True)

End Sub

Private Sub DateKey As String
	
	DateTime.DateFormat = "yyyyMMdd"

	Return HMACSHA256("AWS4" & AWS_Secret_Access_Key, DateTime.Date(AWS_S3_GMT_DateTime), False)
	
End Sub

'************************************************************************************
'
'This procedure creates a HexSHA256Hash of a supplied string
'
'Input parameters are:
'
'       input = string to be mangled
'
'Returns:
'
'       HexSHA256Hash 
'
'Notes on this procedure:
'
'       o HexSHA256Hash is lowercase hexadecimal string of SHA256 hash of input 
'         string
'       o Has some distant parentage at:
'         https://www.b4x.com/android/forum/threads/amazon-s3-library.38699/
'
'************************************************************************************

'
'[Method] - lowercase hexadecimal string of SHA256 hash of input string
'
'input = string to be mangled
Public Sub HexSHA256Hash(input As String) As String
	
	Private wrk_bc As ByteConverter
	Private wrk_md As MessageDigest
	Private wrk_hash() As Byte
	Private wrk_str As String
	
	'Get SHA256 hash of input string
	wrk_hash = wrk_md.GetMessageDigest(input.GetBytes("UTF8"), "SHA-256")
	
	'Get hex of SHA256 hash of input string
	wrk_str = wrk_bc.HexFromBytes(wrk_hash)
	
	'Return lowercase hex of SHA256 hash of input string
	Return wrk_str.ToLowerCase

End Sub

'************************************************************************************
'
'This procedure creates a HMACSHA256 of a supplied string for a given key
'
'Input parameters are:
'
'       key = key
'       input = string to be mangled
'       key_is_hex_string = flag to indicate if key is hex string (True) or 
'                           character string (False)
'
'Returns:
'
'       HMACSHA256
'
'Notes on this procedure:
'
'       o Extracted from:
'         https://www.b4x.com/android/forum/threads/hmacsha1-hash-generation.18600/#post-106868
'
'************************************************************************************

'
'[Method] - lowercase hexadecimal string of HMAC-SHA256 of key and input string
'
'key = key
'input = string to be mangled
'key_is_hex_string = flag to indicate if key is hex string (True) or character string (False)
Public Sub HMACSHA256(key As String, input As String, key_is_hex_string As Boolean) As String

	Private wrk_mac As Mac
	Private wrk_key As KeyGenerator
	Private wrk_bc As ByteConverter
	Private wrk_byte() As Byte
	
	wrk_key.Initialize("HMACSHA256")
	If key_is_hex_string Then
		wrk_key.KeyFromBytes(wrk_bc.HexToBytes(key))
	Else
		wrk_key.KeyFromBytes(key.GetBytes("UTF8"))
	End If
	wrk_mac.Initialise("HMACSHA256", wrk_key.Key)
	wrk_mac.Update(input.GetBytes("UTF8"))
	wrk_byte = wrk_mac.Sign
	
	Return wrk_bc.HexFromBytes(wrk_byte).ToLowerCase
	
End Sub

'************************************************************************************
'
'This procedure URI-encodes an input string, optionally encoding any "/"
'
'Input parameters are:
'
'       input = string to be URI-encoded
'       encodeSlash = flag to indicate if "/" to be encoded
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o This is a port of following java which is at:
'         http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html 
'
'       Public static String UriEncode(CharSequence input, boolean encodeSlash) {
'           StringBuilder result = new StringBuilder();
'           For (int i = 0; i < input.length(); i++) {
'               char ch = input.charAt(i);
'               If ((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || (ch >= '0' && ch <= '9') || ch == '_' || ch == '-' || ch == '~' || ch == '.') {
'                   result.append(ch);
'               } else if (ch == '/') {
'                   result.append(encodeSlash ? "%2F" : ch);
'               } Else {
'                   result.append(toHexUTF8(ch));
'               }
'           }
'           Return result.toString();
'       }
'
'************************************************************************************
Private Sub UriEncode(input As String, encodeSlash As Boolean) As String

	Private i As Int
	Private ch As String
	Private result As String
	
	For i = 0 To input.Length - 1

		ch = input.CharAt(i)
		
		If ((Asc(ch) >= Asc("A") And Asc(ch) <= Asc("Z")) Or (Asc(ch) >= Asc("a") And Asc(ch) <= Asc("z")) Or (Asc(ch) >= Asc("0") And Asc(ch) <= Asc("9")) Or Asc(ch) = Asc("_") Or Asc(ch) = Asc("-") Or Asc(ch) = Asc("~") Or Asc(ch) = Asc(".")) Then

			result = result & ch
			
		Else If ch = "/" Then
			
			If encodeSlash Then
				
				result = result & "%2F"
				
			Else
				
				result = result & ch
				
			End If
			
		Else
			
			result = result & "%" & Bit.ToHexString(Asc(ch)).ToUpperCase
			
		End If
			
	Next
	
	Return result

End Sub

'************************************************************************************
'
'This procedure sorts a map by ascending order of its key
'
'Input parameters are:
'
'       Input_map = map to be sorted
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       None
'
'************************************************************************************
Private Sub SortMap(Input_Map As Map) As Map

	Private wrk_map As Map
	Private wrk_list As List
	Private wrk_ptr As Int
	
	wrk_list.Initialize
	
	For wrk_ptr = 0 To Input_Map.Size - 1
		
		Private wrk_data As KeyValueType
		
		wrk_data.key = Input_Map.GetKeyAt(wrk_ptr)
		wrk_data.value = Input_Map.GetValueAt(wrk_ptr)
		
		wrk_list.Add(wrk_data)

	Next

	wrk_list.SortTypeCaseInsensitive("key", True)
	
	wrk_map.Initialize
		
	Private wrk_data As KeyValueType
	
	For wrk_ptr = 0 To Input_Map.Size - 1
		
		wrk_data = wrk_list.get(wrk_ptr)
		
		wrk_map.Put(wrk_data.key, wrk_data.value)

	Next
	
	Return wrk_map

End Sub

'************************************************************************************
'
'This procedure creates request Signature as per:
'    http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
'    Task 3: Calculate Signature
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       Signature
'
'Notes on this procedure:
'
'       Signature = HMAC-SHA256(SigningKey, StringToSign)
'
'************************************************************************************
Public Sub Signature As String
	
	Return HMACSHA256(SigningKey, StringToSign, True)

End Sub

